
#include"voiceman.h"
#include"ConfigFile.h"

std::string readTextFile(const std::string& fileName);//FIXME:

/**\brief Reads all files from specified directory
 *
 * This function reads all files and returnes their data 
 * as single string. List of files to read is ordered alphabetically previously before reading.
 * This function has no any error indication and files will be silently skipped if they are could not be read.
 *
 * \param [in] dirName Name of a directory to read files from
 *
 * \return Collected files content 
 */
static std::string collectFiles(const std::string& dirName)
{
  StringVector items;
  DIR* dir = opendir(dirName.c_str());
  if (dir == NULL)
    return std::string();
  struct dirent* ent;
  while ((ent = readdir(dir)) != NULL)
    {
      std::string name(ent->d_name);
      if (name == "." || name == "..")
	continue;
      struct stat s;
      std::string nameToCheck;
      if (dirName.empty() || dirName[dirName.length() - 1] == '/')
	nameToCheck = dirName + name; else
	nameToCheck = dirName + "/" + name;
      if (stat(nameToCheck.c_str(), &s) != 0)
	continue;
      if (!S_ISREG(s.st_mode))
	continue;
      items.push_back(name);
    }
  closedir(dir);
  std::string res;
  std::sort(items.begin(), items.end());
  for(StringVector::size_type i = 0;i < items.size();i++)
    {
      std::string nameToRead;
      if (dirName.empty() || dirName[dirName.length() - 1] == '/')
	nameToRead = dirName + items[i]; else
	nameToRead = dirName + "/" + items[i];
      char c;
      std::string s;
      std::ifstream f(nameToRead.c_str());
      if (!f)
	continue;
      while(f.get(c))
	s += c;
      s += '\n';
      res += s;
    } //for files;
  return res;
}

/**\brief Constructs a directory file name to look for extension files
 *
 * this function simply replaces the file extension with ".d" string.
 * If the specified string does not contain dot character ".d" string is added to the end.
 *
 * \param [in] baseName The path to process
 *
 * \return The constructed string
 */
static std::string getExtensionsDirName(const std::string& baseName)
{
  bool dotFound = 0;
  std::string::size_type lastDotPos = 0;
  std::string::size_type i;
  for(i = 0;i < baseName.length();i++)
    {
      if (baseName[i] == '.')
	{
	  lastDotPos = i;
	  dotFound = 1;
	} else
	if (baseName[i] == '/')
	  dotFound = 0;
    }
  if (!dotFound)
    return baseName + ".d";
  std::string res;
  for(i = 0;i < lastDotPos;i++)
    res += baseName[i];
  return res + ".d";
}

ConfigFileSectionVector::size_type ConfigFile::getSectionCount() const
{
  return m_sections.size();
}

const ConfigFileSection& ConfigFile::getSection(ConfigFileSectionVector::size_type index) const
{
  assert(index < m_sections.size());
  return m_sections[index];
}

const ConfigFileSection& ConfigFile::findSection(const std::string& name) const
{
  for(ConfigFileSectionVector::size_type i = 0;i < m_sections.size();i++)
    if (m_sections[i].getName() == toLower(trim(name)))
      return m_sections[i];
  assert(0);
  return m_sections.front();
}

bool ConfigFile::hasSection(const std::string& name) const
{
  for(ConfigFileSectionVector::size_type i = 0;i < m_sections.size();i++)
    if (m_sections[i].getName() == toLower(trim(name)))
      return 1;
  return 0;
}

void ConfigFile::load(const std::string& fileName)
{
  std::string s = readTextFile(fileName);
  s += collectFiles(getExtensionsDirName(fileName));
  StringVector lines;
  std::string ss;
  for(std::string::size_type i = 0;i < s.length();i++)
    {
      if (s[i] == 13)
	continue;
      if (s[i] == 10)
	{
	  lines.push_back(ss);
	  ss.erase();
	  continue;
	}
      ss += s[i];
    }// lines splitting;
  lines.push_back(ss);
  ConfigFileSection sec;
  bool secInited = 0;
  for(StringVector::size_type i = 0;i < lines.size();i++)
    {
      std::string s1, s2;
      int typeCode;
      int resultCode = process(lines[i], typeCode, s1, s2);
      if (resultCode != 0)
	throw ConfigFileException(resultCode, fileName, i + 1);
      if (typeCode == 0)
	continue;
      if (typeCode == 1)
	{
	  if (secInited)
	    m_sections.push_back(sec);
	  secInited = 1;
	  if (s1.empty())
	    throw ConfigFileException(ConfigFileException::EMPTY_SECTION_NAME, fileName, i + 1);
	  sec.setName(s1);
	  sec.clear();
	  continue;
	} // new section;
      if (typeCode == 2)
	{
	  if (!secInited)
	    throw ConfigFileException(ConfigFileException::PARAMETER_WITHOUT_SECTION, fileName, i + 1);
	  sec.add(trim(s1), s2);
	  continue;
	} // param;
      assert(0);
    } // for lines;
  if (secInited)
    m_sections.push_back(sec);
}

/*
 * This method parses configuration file line. It returns error code or
 * zero if line was successfully parsed. An error code, returned by this
 * method can be passed to ConfigFileException object. 
 *
 * Through the code parameter this method returns determined line
 * type. The following values are used:
 *  0 - it is an empty line and must be ignored;
 *  1 - it is a section header line;
 *  2 - it is parameter value line.
 */
int ConfigFile::process(const std::string& line, int& code, std::string& str1, std::string& str2) const
{
  const std::string loLine = toLower(line);
  int state = 0;
  str1.erase();
  str2.erase();
  code = 0;
  for(std::string::size_type i = 0;i < line.length();i++)
    {
      if (state == 0 && BLANK_CHAR(line[i]))
	continue;
      if (state == 0 && line[i] == '#')
	return 0;
      if (state == 0 && line[i] == '[')
	{
	  code = 1;//it is section header;
	  for(i++;i < line.length();i++)
	    {
	      if (BLANK_CHAR(line[i]))
		continue;
	      if (line[i] == ']')
		return 0;
	      if (loLine[i] >= 'a' && loLine[i] <= 'z')
		{
		  str1 += loLine[i];
		  continue;
		}
	      return ConfigFileException::INVALID_SECTION_HEADER;
	    }  // for;
	} // section;
      if (state == 0 && loLine[i] >= 'a' && loLine[i] <= 'z')
	{
	  str1 += loLine[i];
	  code = 2;
	  state = 1;
	  continue;
	}
      if (state == 1 && (BLANK_CHAR(line[i]) || line[i]=='_'))
	continue;
      if (state == 1 && loLine[i] >= 'a' && loLine[i] <= 'z')
	{
	  str1 += loLine[i];
	  continue;
	}
      if (state == 1 && DIGIT_CHAR(line[i]))
	{
	  str1 += line[i];
	  continue;
	}
      if (state == 1 && line[i] == '=')
	{
	  state = 2;
	  continue;
	}
      if (state == 2 && BLANK_CHAR(line[i]))
	continue;
      if (state == 2 && ((loLine[i] >= 'a' && loLine[i] <= 'z') || DIGIT_CHAR(line[i])))
	{
	  for(;i < line.length();i++)
	    {
	      if (line[i] == '_' || (loLine[i] >= 'a' && loLine[i] <= 'z') || DIGIT_CHAR(line[i]))
		str2 += line[i]; else
		  break;
	    } // for;
	  for(;i < line.length();i++)
	    {
	      if (BLANK_CHAR(line[i]))
		continue;
	      if (line[i] == '#')
		return 0;
	      return ConfigFileException::INVALID_UNQUOTED_VALUE;
	    }
	  return 0;
	} // no quotas;
      if (state == 2 && line[i] == '\"')
	{
	  state = 3;
	  continue;
	}
      if (state == 3 && line[i] != '\"')
	{
	  str2 += line[i];
	  continue;
	}
      if (state == 3 && line[i] == '\"')
	{
	  state = 4;
	  continue;
	}
      if (state == 4 && line[i] == '\"')
	{
	  state = 3;
	  str2 += '\"';
	  continue;
	}
      if (state == 4 && line[i] != '\"')
	return 0;
      switch(state)
	{
	case 0:
	  return ConfigFileException::INVALID_LINE_BEGINNING;
	case 1:
	  return ConfigFileException::INVALID_PARAMETER_NAME;
	case 2:
	  return ConfigFileException::INVALID_CHAR_AFTER_EQUALS;
	default:
	  assert(0);
	}
    } // for;
  assert(state!=0 || code==0);
  if (state == 4 || state == 0)
    return 0;
  return ConfigFileException::UNEXPECTED_LINE_END;;
}

int ConfigFile::params(const std::string& section, const std::string& param) const
{
  return ANY_VALUE;
}

void ConfigFile::checkParamsInSection(const ConfigFileSection& section) const
{
  for(ConfigFileSection::const_iterator it = section.begin();it != section.end();it++)
    {
      int requiredType = params(section.getName(), it->first);
      const std::string& value = it->second;
      if (requiredType == ANY_VALUE)
	continue;
      if (requiredType == INVALID_VALUE)
	throw ConfigFileValueTypeException("unknown parameter \'" + it->first + "\' in section \'" + section.getName() + "\'");
      if (requiredType == BOOLEAN_VALUE && !checkTypeBool(value))
	throw ConfigFileValueTypeException("parameter \'" + it->first + "\' in section \'" + section.getName() + "\' should be boolean. value \'" + trim(value) + "\' is illegal");
      if (requiredType == STRING_VALUE && trim(value).empty())
	throw ConfigFileValueTypeException("parameter \'" + it->first + "\' in section \'" + section.getName() + "\' cannot have an empty value");
      if (requiredType == INT_VALUE && !checkTypeInt(value))
	throw ConfigFileValueTypeException("parameter \'" + it->first + "\' in section \'" + section.getName() + "\' should be an integer number. Value \'" + trim(value) + "\' is illegal");
      if (requiredType == UINT_VALUE && !checkTypeUnsignedInt(value))
	throw ConfigFileValueTypeException("parameter \'" + it->first + "\' in section \'" + section.getName() + "\' should be an unsigned integer number. Value \'" + trim(value) + "\' is illegal");
      if (requiredType == DOUBLE_VALUE && !checkTypeDouble(value))
	throw ConfigFileValueTypeException("parameter \'" + it->first + "\' in section \'" + section.getName() + "\' should be an floating point number. Value \'" + trim(value) + "\' is illegal");
    } //for();
}

void ConfigFile::checkParams() const
{
  for(ConfigFileSectionVector::size_type i = 0;i < m_sections.size();i++)
    checkParamsInSection(m_sections[i]);
}
